"""
图形界面部分
"""

from math import pi
from tkinter import Event, Menu, messagebox

import tkintertools as tkt
from configure import config
from constants import SIDE
from cube import MagicCube
from main import __author__, __update__, __version__


class Window:
    """ 窗口界面 """

    root = tkt.Tk('3D魔方', SIDE, SIDE)
    root.iconbitmap('icon.ico')
    root.resizable(False, False)
    menu = Menu(root, tearoff=False)
    root.configure(menu=menu)
    canvas = tkt.Canvas(root, SIDE, SIDE)
    canvas.place(x=0, y=0)

    def __init__(self) -> None:
        self.init_menu()
        self.init_bind()
        self.init_ui()
        self.event = [None, None]
        self.mcube = MagicCube(self.canvas, 50, config['order'])
        self.root.mainloop()

    def init_menu(self) -> None:
        """ 菜单栏初始化 """
        m1 = Menu(self.menu, tearoff=False)
        m2 = Menu(self.menu, tearoff=False)
        self.menu.add_cascade(menu=m1, label='选项(O)')
        self.menu.add_cascade(menu=m2, label='帮助(H)')
        m1.add_command(label='设置')
        m1.add_command(label='新游戏', accelerator='Ctrl+O')
        m1.add_command(label='随机打乱')
        m1.add_command(label='退出', accelerator='Alt+F4',
                       command=self.root.quit)
        m2.add_command(label='游戏帮助')
        m2.add_command(label='关于', command=self.page_about)

    def init_bind(self) -> None:
        """ 绑定初始化 """
        self.root.bind('<B3-Motion>', self.rotate)
        self.root.bind('<Button-3>', self.record)
        self.root.bind('<MouseWheel>', self.scele)
        self.root.bind('<Any-Key>', self.turn)

    def init_ui(self) -> None:
        """ UI初始化 """
        self.info = self.canvas.create_text(
            SIDE-10, 15, text='x: 0.000 y: 0.000\ns: 100%', justify='right', anchor='e', font=('楷体', 12))

    def page_about(self) -> None:
        """ 关于页面 """
        messagebox.showinfo(
            '3D魔方', '作者：%s\n更新：%s\t\t\n版本：%s' % (__author__, __update__, __version__))

    def record(self, event: Event) -> None:
        """ 记录位置 """
        self.event = [event.x, event.y]

    def scele(self, event: Event) -> None:
        """ 缩放 """
        if event.delta > 0 and MagicCube.scale < 1.49:
            MagicCube.scale += 0.05
        elif event.delta < 0 and MagicCube.scale > 0.51:
            MagicCube.scale -= 0.05
        self.update()

    def rotate(self, event: Event) -> None:
        """ 转动魔方 """
        r_x, r_y = self.event[0]-event.x, self.event[1]-event.y
        self.mcube.rotate(
            0, r_y/config['sensitivity'], -r_x/config['sensitivity'])
        self.event = [event.x, event.y]
        self.update()

    def update(self) -> None:
        """ 更新 """
        self.mcube.update()
        self.canvas.itemconfigure(
            self.info, text='x: %.3f y: %.3f\ns: %d%%' % (self.mcube.x/pi*180, self.mcube.y/pi*180, round(MagicCube.scale*100)))

    def turn(self, event: Event) -> None:
        """ 转动 """
        self.mcube.turn(event.char)
