"""
核心文件：基本3D的运行
"""

from math import atan, cos, pi, sin

import tkintertools as tkt
from configure import config
from constants import SIDE


def arctan(n1: float, n2: float) -> float:
    """ 反正切 """
    if n2:
        return atan(n1/n2)
    elif n1 > 0:
        return pi/2
    return -pi/2


def dot(m1: list[list[float]], m2: list[float, float, float]) -> list[float]:
    """ 矩阵相乘 """
    for i in range(3):
        m1[0][i] = sum(m1[i][j]*m2[j] for j in range(3))
    return m1[0]


class Point:
    """ 点 """

    def __init__(self, x: float, y: float, z: float) -> None:
        self.x = x
        self.y = y
        self.z = z

    def translate(self, dx: float, dy: float, dz: float) -> None:
        """ 三维平移 """
        self.x += dx
        self.y += dy
        self.z += dz

    def rotate(self, dα: float, dβ: float, dγ: float) -> None:
        """ 三维旋转 """
        sa, sb, sc = sin(dα), sin(dβ), sin(dγ)
        ca, cb, cc = cos(dα), cos(dβ), cos(dγ)
        M = [[cc*cb, cc*sb*sa-sc*ca, cc*sb*ca+sc*sa],
             [sc*cb, sc*sb*sa+cc*ca, sc*sb*ca-cc*sa],
             [-sb,   cb*sa,          cb*ca]]
        self.x, self.y, self.z = dot(M, [self.x, self.y, self.z])

    def project(self, dis: float) -> tuple[float, float]:
        """ 返回二维坐标 """
        k = dis / (dis - self.x / config['DoF'])
        return self.y*k, self.z*k


class Side:
    """ 面 """

    def __init__(self, canvas: tkt.tkinter.Canvas, color: str, dx: float, dy: float, *args: tuple[float, float, float]) -> None:
        self.points = [Point(x, y, z) for x, y, z in args]
        self.update_xs()
        self.dx, self.dy = dx, dy
        self.canvas = canvas
        p = [i + (dy if p.index(i) & 1 else dx)
             for p in self.project(config['distance']) for i in p]
        self.polygon = canvas.create_polygon(
            *p, fill=color, outline='black')

    def translate(self, dx: float, dy: float, dz: float) -> None:
        """ 三维平移 """
        for point in self.points:
            point.translate(dx, dy, dz)
        self.update_xs()

    def rotate(self, dx: float, dy: float, dz: float) -> None:
        """ 三维旋转 """
        for point in self.points:
            point.rotate(dx, dy, dz)
        self.update_xs()

    def update_xs(self) -> None:
        """ 更新self.xs """
        self.xs = min(point.x for point in self.points)

    def project(self, dis: float) -> tuple[tuple[float, float]]:
        """ 二维投影 """
        return tuple(point.project(dis) for point in self.points)

    def update(self) -> None:
        """ 更新图像 """
        self.update_xs()
        p = [i * MagicCube.scale + (self.dy if p.index(i) & 1 else self.dx)
             for p in self.project(config['distance']) for i in p]
        self.canvas.coords(self.polygon, *p)


class Cube:
    """ 正方体 """

    def __init__(self, x: float, y: float, z: float, dx: float, dy: float, r: float, canvas: tkt.Canvas) -> None:
        self.canvas = canvas
        p = [(x+a, y+b, z+c) for a in (r, -r)
             for b in (r, -r) for c in (r, -r)]
        self.sides = [
            Side(canvas, 'orange', dx, dy, p[4], p[5], p[7], p[6]),  # 后面
            Side(canvas, 'blue',   dx, dy, p[2], p[3], p[7], p[6]),  # 左侧
            Side(canvas, 'yellow', dx, dy, p[1], p[3], p[7], p[5]),  # 上面
            Side(canvas, 'green',  dx, dy, p[0], p[1], p[5], p[4]),  # 右侧
            Side(canvas, 'white',  dx, dy, p[0], p[4], p[6], p[2]),  # 下面
            Side(canvas, 'red',    dx, dy, p[0], p[1], p[3], p[2])]  # 前面
        self.update_xs()

    def rotate(self, dx: float, dy: float, dz: float) -> None:
        """ 旋转 """
        for side in self.sides:
            side.rotate(dx, dy, dz)
        self.update_xs()

    def translate(self, dx: float, dy: float, dz: float) -> None:
        """ 平移 """
        for side in self.sides:
            side.translate(dx, dy, dz)
        self.update_xs()

    def update(self) -> None:
        """ 更新立方体 """
        self.update_xs()
        for side in self.sides:
            side.update()
        self.lift()

    def update_xs(self) -> None:
        """ 更新self.xs """
        self.xs = min(side.xs for side in self.sides)

    def lift(self) -> None:
        """ 前后关系 """
        for side in sorted(self.sides, key=lambda s: s.xs):
            self.canvas.lift(side.polygon)


class MagicCube:
    """ 魔方类 """

    scale = 1  # 缩放系数

    def __init__(self, canvas: tkt.Canvas, r: float, n: int) -> None:
        self.n = n
        self.r = r
        k = [4*r*(i-(n-1)/2)/(n-1) for i in range(n)]
        self.x, self.y = 0, 0
        self.cubes = [Cube(a, b, c, SIDE/2, SIDE/2, 2*r/(n-1), canvas)
                      for a in k for b in k for c in k]

    def rotate(self, dx: float, dy: float, dz: float) -> None:
        """ 旋转 """
        self.x += dz
        self.y += dy
        for cube in self.cubes:
            cube.rotate(dx, dy, dz)

    def translate(self, dx: float, dy: float, dz: float) -> None:
        """ 平移 """
        for cube in self.cubes:
            cube.translate(dx, dy, dz)

    def update(self) -> None:
        """ 更新 """
        for cube in self.cubes:
            cube.update()
        self.lift()

    def lift(self) -> None:
        """ 更新前后关系 """
        for cube in sorted(self.cubes, key=lambda c: c.xs):
            cube.lift()

    def turn(self, key: str) -> None:
        """ 转动 """
        [range(9), range(9, 18), range(18, 27)]
        [range(0, 27, 3), range(1, 27, 3), range(2, 27, 3)]
        [0, 1, 2, 9,  10, 11, 18, 19, 20]
        [1, 2, 3, 10, 11, 12, 19, 20, 21]
        [2, 3, 4, 11, 12, 13, 20, 21, 22]
        if key == 'r':
            for i in [0, 1, 2, 9,  10, 11, 18, 19, 20]:
                self.cubes[i].rotate(0, pi/2, 0)
        self.update()
